import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
import time
from scipy.spatial.distance import pdist, squareform
from numpy.linalg import norm
import cv2 as cv

class Calculater():
    def __init__(self, img, edge):
        self.img = img.astype(float)
        self.edge = edge

        self.find_convexhull()
        self.y0, self.x0 = ndimage.center_of_mass(self.hull)

    def find_convexhull(self):
        # Method calculates convexhull.
        # If edge flag is set, it uses the corners not centers
        if self.edge:
            ys, xs = np.nonzero(self.img)
            new_xs = np.concatenate(
                (xs+0.5, xs+0.5, xs-0.5, xs-0.5, xs, xs+0.5, xs-0.5, xs, xs))
            new_ys = np.concatenate(
                (ys+0.5, ys-0.5, ys+0.5, ys-0.5, ys, ys, ys, ys+0.5, ys-0.5))
            new_ys, new_xs = (new_ys*2).astype(int), (new_xs*2).astype(int)
            self.hull = cv.convexHull(np.array([new_ys, new_xs]).T).T.reshape(2, -1).astype(float)
        else:
            self.hull = cv.convexHull(np.transpose(np.nonzero(self.img))).T.reshape(2, -1).astype(float)


    def plot(self):
        """
        This method plots the particle, the maxferet and minferet coordiantes
        and the lines which run through the maxferet and minferet.

        """
        ymax, xmax = self.img.shape

        xs = np.linspace(0, xmax, 2)

        minf_ys = np.tan(self.minf_angle) * xs + self.minf_t
        maxf_ys = np.tan(self.maxf_angle) * xs + self.maxf_t

        minf_base_m = (self.minf_coords.T[0][0]-self.minf_coords.T[0][1])/(self.minf_coords.T[1][0]-self.minf_coords.T[1][1])
        minf_base_t = self.minf_coords.T[0][0] - minf_base_m*self.minf_coords.T[1][0]
        minf_anker_t = self.minf_coords.T[0][2] - minf_base_m*self.minf_coords.T[1][2]

        minf_base_ys = minf_base_m*xs + minf_base_t
        minf_anker_ys = minf_base_m*xs + minf_anker_t

        maxf_base_t = self.maxf_coords.T[0][0] - np.tan(self.maxf_angle + np.pi/2) * self.maxf_coords.T[1][0]
        maxf_anker_t = self.maxf_coords.T[0][1] - np.tan(self.maxf_angle + np.pi/2) * self.maxf_coords.T[1][1]

        maxf_base_ys = np.tan(self.maxf_angle + np.pi/2)* xs + maxf_base_t
        maxf_anker_ys = np.tan(self.maxf_angle + np.pi/2)* xs + maxf_anker_t


        plt.figure(dpi=100, figsize=(10, 10))
        plt.title(f'MinFeret: {self.minf:.6f} ||| MaxFeret: {self.maxf:.6f}')
        plt.imshow(self.img, origin='lower')

        plt.scatter(self.maxf_coords.T[1], self.maxf_coords.T[0], label='MaxFeret Coordinates')
        plt.plot(xs, maxf_ys, color='green', label='Maxferet Line')
        plt.plot(xs, maxf_base_ys, linestyle='--', color='green', label='MaxFeret Baseline')
        plt.plot(xs, maxf_anker_ys, linestyle='--', color='green')

        plt.scatter(self.minf_coords.T[1], self.minf_coords.T[0], label='MinFeret Coordinates')
        plt.plot(xs, minf_ys, color='orange', label='Minferet Line')
        plt.plot(xs, minf_base_ys, linestyle='--', color='orange', label='MinFeret Baseline')
        plt.plot(xs, minf_anker_ys, linestyle='--', color='orange')
        plt.ylim(0, ymax)
        plt.xlim(0, xmax)
        plt.legend()
        plt.show()



    def calculate_minferet(self):
        """
        Method calculates the exact minimum feret diameter.
        The result is equal to imagejs minferet.

        """
        length = len(self.hull.T)

        Ds = np.empty(length)
        ps = np.empty((length, 3, 2))


        for i in range(length):
            p1 = self.hull.T[i]
            p2 = self.hull.T[(i+1) % length]
            
            ds = np.abs(np.cross(p2-p1, p1-self.hull.T)/norm(p2-p1))

            Ds[i] = np.max(ds)

            d_i = np.where(ds == Ds[i])[0][0]
            p3 = self.hull.T[d_i]
            ps[i] = np.array((p1, p2, p3))


        self.minf = np.min(Ds)

        minf_index = np.where(Ds == self.minf)[0][0]

        (y0, x0), (y1, x1), (y2, x2) = ps[minf_index]

        m = (y0 - y1) / (x0 - x1)
        t = y0 - m * x0

        self.minf_coords = np.array(((y0, x0), (y1, x1), (y2, x2))) 

        self.minf_angle = np.arctan(m) + np.pi/2

        if self.minf_angle < 0:
            self.minf_angle += np.pi

        if self.edge:
            self.minf /= 2.
            self.minf_coords /= 2.

        self.minf_t = self.minf_coords.T[0][2] - np.tan(self.minf_angle) * self.minf_coords.T[1][2]
        


    def calculate_maxferet(self):
        """
        The maxferet is defined as the maximum euclidean
        distance between two points. pdist calculates
        all the distances between the points and than
        the maximum is taken from all those.

        """
        pdists = pdist(self.hull.T, "euclidean")

        self.maxf = np.max(pdists)

        maxf_coords_index = np.where(squareform(pdists) == self.maxf)[0]

        # If there are more then one maxferet-combination, this two lines
        # sort them in x and y and chooses the first element.
        maxf_coords_index_y = maxf_coords_index[:len(maxf_coords_index)//2][0]
        maxf_coords_index_x = maxf_coords_index[len(maxf_coords_index)//2:][0]

        self.maxf_coords = self.hull.T[np.array((maxf_coords_index_x, maxf_coords_index_y))]


        ((y0, x0), (y1, x1)) = self.maxf_coords

        m = (y0 - y1) / (x0 - x1)
        t = y0 - m * x0

        self.maxf_angle = np.arctan(m)



        if self.maxf_angle < 0:
            self.maxf_angle += np.pi


        if self.edge:
            self.maxf /= 2.
            self.maxf_coords /= 2

        self.maxf_t = self.maxf_coords.T[0][1] - np.tan(self.maxf_angle) * self.maxf_coords.T[1][1]



    def calculate_maxferet90(self):
        """
        Method calculates the feret diameter which
        is 90 degree to the maximum feret diameter.
        It first checks if the angle of the maximum
        feret diameter is already calculatet. If not
        it calls the maxferet function.

        """
        self.maxf90_angle =self.maxf_angle + np.pi/2
        self.maxf90 = self.calculate_distances(self.maxf90_angle)
        
        if self.edge:
            self.maxf90 /= 2


    def calculate_minferet90(self):
        """
        Method calculates the feret diameter which
        is 90 degree to the minimum feret diameter.
        It first checks if the angle of the minimum
        feret diameter is already calculatet. If not
        it calls the minferet function.

        """        
        self.minf90_angle =self.minf_angle + np.pi/2
        self.minf90 = self.calculate_distances(self.minf90_angle)

        if self.edge:
            self.minf90 /= 2


    def calculate_distances(self, a):
        """ 
        Method calculates the distance of two points at a givin angle.

        Args:
            a (float): angle (in rad)

        Returns:
            distance (float): caliper distance for angle a
        """
        m = np.tan(a)
        ds = np.cos(a)*(self.y0-self.hull[0])-np.sin(a)*(self.x0-self.hull[1])
        max_i = np.argmax(ds)
        min_i = np.argmin(ds)

        t_max = self.hull.T[max_i][0] - m * self.hull.T[max_i][1]
        t_min = self.hull.T[min_i][0] - m * self.hull.T[min_i][1]

        distance = np.abs(t_max - t_min) / np.sqrt(1+m**2)

        return distance


# 
# from main import Calculater

from feret.main import Calculater


def calc(img, edge=False):
    """
    Calculate the true maximum feret diameter (minf)
    and the approximated minimum feret diameter (minf).

    result = calc(img)
    maxf, minf = result.maxf, result.minf

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        Calculater (object): calculator class.
    """
    feret_calc = Calculater(img, edge)
    feret_calc.calculate_minferet()
    feret_calc.calculate_minferet90()
    feret_calc.calculate_maxferet()
    feret_calc.calculate_maxferet90()
    return feret_calc


def all(img, edge=False):
    """
    Calculate the true maximum feret diameter (minf)
    and the approximated minimum feret diameter (minf).

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        maxf (float): maximum feret diameter
        minf (float): minimum feret diameter
    """

    feret_calc = calc(img, edge)
    return feret_calc.maxf, feret_calc.minf, feret_calc.minf90, feret_calc.maxf90


def plot(img, edge=False):
    """
    Plot of the results.

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    """
    feret_calc = calc(img, edge)
    feret_calc.plot()


def max(img, edge=False):
    """
    Calculate the true maximum feret diameter (minf).

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        maxf (float): maximum feret diameter
    """

    feret_calc = Calculater(img, edge)
    feret_calc.calculate_maxferet()
    return feret_calc.maxf


def min(img, edge=False):
    """
    Calculate the approximated minimum feret diameter (minf).

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        minf (float): minimum feret diameter
    """

    feret_calc = Calculater(img, edge)
    feret_calc.calculate_minferet()
    return feret_calc.minf


def min90(img, edge=False):
    """
    Calculate the approximated feret diameter 90  
    degree (minf90) to minimum feret diameter (minf).

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        minf90 (float): minimum feret diameter 90 degree
    """

    feret_calc = Calculater(img, edge)
    feret_calc.calculate_minferet()
    feret_calc.calculate_minferet90()
    return feret_calc.minf90


def max90(img, edge=False):
    """
    Calculate the approximated feret diameter 90  
    degree (maxf90) to maximum feret diameter (minf).

    Args:
        img (numpy.ndarray): binary-image
        edge (boolean): use edges (vertices) or centers

    Returns:
        maxf90 (float): maximum feret diameter 90 degree
    """

    feret_calc = Calculater(img, edge)
    feret_calc.calculate_maxferet()
    feret_calc.calculate_maxferet90()
    return feret_calc.maxf90


# if __name__ == '__main__':

#     import numpy as np
#     import time
#     import tifffile as tif
#     # img = np.load('img.npy')
#     img = tif.imread('11300_binary_verrgroessert.tif').T
#     img = tif.imread('126_binary_verrgroessert.tif')
    

#     res = plot(img, edge=True)

#     print(res.maxf_coords)






# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['feret']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'opencv-python>=4.5,<5.0',
 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'feret',
    'version': '1.2.7',
    'description': 'Calculate the maximum and minimum Feret diameter.',
    'long_description': '# *Feret*: A Python Module to calculate the Feret Diameter of Binary Images\n\n<img src="docs\\plot_method.png" style="zoom: 33%;" />\n\nThis python module can calculate the following parameters for binary images:\n\n* maximum Feret diameter (maxferet, maxf)\n* minimum Feret diameter (minferet, minf)\n* Feret diameter 90 ° to the minferet (minferet90, minf90) \n* Feret diameter 90 ° to maxferet (maxferet90, maxf90) \n\nSee this [Wikipedia page](https://en.wikipedia.org/wiki/Feret_diameter) to get the definition of those parameters.\n\nThis module gives the exact results as ImageJ (use `edge=True` as shown below), all the parameters are exactly calculated and **not** approximated.\n\n## Installations\nThis project is available via pip:\n\n`pip install feret`\n\n## Pieces of Information\n\n#### Convex Hull\n\nThe definition of the maxferet and minferet uses the image of a caliper. Therefore, only the points which correspond to the convex hull of the object play a role. That is why before any calculations the convex hull is determined to reduce the runtime.\n\n#### Maxferet\nThe maxferet is calculated as the maximum Euclidean distance of all pixels.\n\n#### Minferet\n\nThe minferet is exactly calculated and **not** approximated. My algorithm uses the fact, that the calipers that define the minferet run on one side through two points and on the other through one point. The script iterates over all edge points and defines a line through the point and the one next to it. Then all the distances to the other points are calculated and the maximum is taken. The minimum of all those maximums is the minferet. The maximum of all those maximums is **not** the maxferet, that is the reason it is calculated separately. The runtime of this is already pretty good but hopefully I can improve it in the future.\n\n## Use\nThe module can be used as followed:\n\nFirst you need a binary image for which the feret diameter should be calculated. The background has to have the value zero, the object can have any nonzero  value. The object doesn\'t have to be convex. At the moment the module only supports one object per image.This means, that if there are multiple not connected regions, the script will calculate a convexhull which include all regions and for this hull the feret diameter is calculated.\n\nThr calls are:\n\n```python\nimport feret\n\n# tifffile is not required nor included in this module.\nimport tifffile as tif\nimg = tif.imread(\'example.tif\') # Image has to be a numpy 2d-array.\n\n\n# get the values\nmaxf, minf, minf90, maxf90 = feret.all(img)\n\n# get only maxferet\nmaxf = feret.max(img)\n\n# get only minferet\nminf = feret.min(img)\n\n# get only minferet90\nminf90 = feret.min90(img)\n\n# get only maxferet90\nmaxf90 = feret.max90(img)\n\n# get all the informations\nres = feret.calc(img)\nmaxf = res.maxf\nminf =  res.minf\nminf90 = res.minf90\nminf_angle = res.minf_angle\nminf90_angle = res.minf90_angle\nmaxf_angle = res.maxf_angle\nmaxf90_angle = res.maxf90_angle\n```\n\nThere is an option to calculate the Feret diameters for the pixel edges instead of the centers. Just add an `edge=True` in the call as shown below. This works for all calls analogous.\n\n```python\nimport feret\n\n# tifffile is not required nor included in this module.\nimport tifffile as tif\nimg = tif.imread(\'example.tif\') # Image has to be a numpy 2d-array.\n\n# get only maxferet\nmaxf = feret.max(img, edge=True)\n```\n\nThis module can also plot the result. Just use\n\n```python\nimport feret\n\n# tifffile is not required nor included in this module.\nimport tifffile as tif\nimg = tif.imread(\'example.tif\') # Image has to be a numpy 2d-array.\n\n# plot the result\nferet.plot(img) #edge=True can be passed here too\n```\n\n<img src="docs\\plot_method.png" style="zoom: 33%;" />\n\nThe reason for the two MinFeret points on the left is described above. The MinFeret line does not have to run in between its two base points or through one of them. MaxFeret and MinFeret do not have to be 90° to each other. To calculate the 90° to MaxFeret and MinFeret use `feret.max90(img)` and `feret.min90(img)` methods.\n',
    'author': 'matthiasnwt',
    'author_email': '62239991+matthiasnwt@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/matthiasnwt/feret',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3,<4',
}


setup(**setup_kwargs)
